/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      ClassInfo
Version:   1.0
Status:    Beta
ID:        ID_CLASSINFO
Date:      August 2004
Category:  System
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2004.  All rights reserved.
Short:     Provides support for scanning classes and objects in the system.
Keywords:  class, object, info, information, scan, search

###DESCRIPTION###
<p>The ClassInfo class is provided for object script developers that need to
be able to scan the system for classes and retrieve information on the objects
that they have allocated.</p>



###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_CLASSINFO 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_CLASSINFO, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2004.  All rights reserved.",
   "August 2004",
   "ClassInfo"
};

static struct KernelBase  *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static OBJECTPTR modStrings = 0;
static OBJECTPTR ClassInfoClass = 0;

/*****************************************************************************
** Class definition.
*/

struct ClassInfo {
   OBJECT_HEADER
   OBJECTID *Objects;
   CLASSID  ClassID;
   LONG     TotalObjects;
   LONG     Flags;

   PRIVATE_FIELDS
   UBYTE prvClass[40];
};

static ERROR GET_Class(struct ClassInfo *, STRING *);

static ERROR SET_Class(struct ClassInfo *, STRING);
static ERROR SET_ClassID(struct ClassInfo *, LONG *);

#define CIF_PUBLIC  0x00000001
#define CIF_PRIVATE 0x00000002

static struct FieldDef InfoFlags[] = {
   { "PUBLIC",  CIF_PUBLIC },
   { "PRIVATE", CIF_PRIVATE },
   { NULL, NULL }
};

static struct FieldArray ClassInfoFields[] = {
   { "Objects",      0, FDF_POINTER|FDF_LONG|FDF_R, 0, NULL, NULL },
   { "ClassID",      0, FDF_LONG|FDF_RW,            0, NULL, SET_ClassID },
   { "TotalObjects", 0, FDF_LONG|FDF_R,             0, NULL, NULL },
   { "Flags",        0, FDF_LONG|FDF_RI,            (LONG)&InfoFlags, NULL, NULL },
   /*** Virtual fields ***/
   { "Class",   0, FDF_STRING|FDF_RW,  0, GET_Class, SET_Class },
   END_FIELD
};

static ERROR INFO_Free(struct ClassInfo *, APTR);
static ERROR INFO_GetUnlistedField(struct ClassInfo *, struct acGetUnlistedField *);
static ERROR INFO_Init(struct ClassInfo *, APTR);
static ERROR INFO_Refresh(struct ClassInfo *, APTR);

static struct ActionArray ClassInfoActions[] = {
   { AC_Free,             INFO_Free },
   { AC_GetUnlistedField, INFO_GetUnlistedField },
   { AC_Init,             INFO_Init },
   { AC_Refresh,          INFO_Refresh },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ClassInfoClass, NULL,
      FID_BaseClassID|TLONG, ID_CLASSINFO,
      FID_Version|TFLOAT,    VER_CLASSINFO,
      FID_Name|TSTRING,      "ClassInfo",
      FID_Category|TLONG,    CCF_SYSTEM,
      FID_Actions|TPTR,      ClassInfoActions,
      FID_Fields|TPTR,       ClassInfoFields,
      FID_Size|TLONG,        sizeof(struct ClassInfo),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (modStrings)     { acFree(modStrings);     modStrings     = NULL; }
   if (ClassInfoClass) { acFree(ClassInfoClass); ClassInfoClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** ClassInfo: Free
*/

static ERROR INFO_Free(struct ClassInfo *Self, APTR Void)
{
   if (Self->Objects) {
      FreeMemory(Self->Objects);
      Self->Objects = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  GetUnlistedField
Short: Support is provided for unlisted fields.

###DESCRIPTION###
<p>Unlisted fields are supported by the ClassInfo class in order to aid script
usage.</p>

<p>To retrieve the ID of any object in the Objects array, use a field name of
"object(index)" where 'index' is the array index that you are interested in.
You can use this feature to scan the Objects array within a repeat loop, for
example:</p>

<pre>
   &lt;repeat count="[classinfo.totalobjects]"&gt;
     &lt;xml&gt;
       &lt;item&gt;[classinfo.object({index})]&lt;/item&gt;
     &lt;/xml&gt;
   &lt;/repeat&gt;
</pre>

###END###

*****************************************************************************/

static ERROR INFO_GetUnlistedField(struct ClassInfo *Self, struct acGetUnlistedField *Args)
{
   STRING buffer;

   buffer = Args->Buffer;
   buffer[0] = 0;

   if (StrCompare("object(", Args->Field, 7, NULL) IS ERR_Okay) {
      WORD index;

      index = StrToInt(Args->Field+7);

      if ((index < 0) OR (index > Self->TotalObjects)) return(ERR_OutOfRange);
      if (!Self->Objects) return(ERR_FieldNotSet);

      IntToStr(Self->Objects[index], buffer, Args->Size);

      return(ERR_Okay);
   }
   else return(ERR_NoSupport);
}

/*****************************************************************************
** ClassInfo: Init
*/

static ERROR INFO_Init(struct ClassInfo *Self, APTR Void)
{
   INFO_Refresh(Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Refresh
Short: Refreshes the list of objects associated with the class.
###END###

*****************************************************************************/

static ERROR INFO_Refresh(struct ClassInfo *Self, APTR Void)
{
   WORD i, j;

   if (Self->Objects) {
      FreeMemory(Self->Objects);
      Self->Objects = NULL;
   }

   if (!Self->ClassID) return(ERR_Okay);

   if (FindObject(NULL, Self->ClassID, &Self->Objects, &Self->TotalObjects) IS ERR_Okay) {
      if ((Self->Flags & CIF_PUBLIC) AND (Self->Flags & CIF_PRIVATE)) {
         /*** Do nothing as we already have private and public objects ***/
      }
      else if (Self->Flags & CIF_PUBLIC) {
         /*** We are only interested in public objects ***/
         for (i=0,j=0; i < Self->TotalObjects; i++) {
            if (Self->Objects[i] > 0) continue;
            Self->Objects[j++] = Self->Objects[i];
         }
         Self->Objects[j] = 0;
         Self->TotalObjects = j;
      }
      else if (Self->Flags & CIF_PRIVATE) {
         /*** We are only interested in private objects ***/
         for (i=0,j=0; i < Self->TotalObjects; i++) {
            if (Self->Objects[i] < 0) continue;
            Self->Objects[j++] = Self->Objects[i];
         }
         Self->Objects[j] = 0;
         Self->TotalObjects = j;
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Class
Short:  The name of the class to be targetted.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The name of the class that you want to target for examination must be
indicated here.  If this field is set after the classinfo object has been
initialised, the Objects array will be refreshed immediately.</p>

###END###

*****************************************************************************/

static ERROR GET_Class(struct ClassInfo *Self, STRING *Value)
{
   *Value = Self->prvClass;

   if ((Self->ClassID) AND (!Self->prvClass[0])) {
      struct SysClass *class;
      if ((class = (struct SysClass *)FindClass(Self->ClassID, NULL))) {
         StrCopy(class->Name, Self->prvClass, sizeof(Self->prvClass));
      }
   }

   return(ERR_Okay);
}

static ERROR SET_Class(struct ClassInfo *Self, STRING Value)
{
   WORD i;

   if ((!Value) OR (!*Value)) {
      Self->prvClass[0] = 0;
      Self->ClassID = 0;
      return(ERR_Okay);
   }

   for (i=0; (Value[i]) AND (i < sizeof(Self->prvClass)-1); i++) Self->prvClass[i] = Value[i];
   Self->prvClass[i] = 0;

   Self->ClassID = ResolveClassName(Self->prvClass);

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      acRefresh(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ClassID
Short:  The ID of the class to be targetted.
Type:   STRING
Status: Read/Set

###DESCRIPTION###
<p>The ID of the class that you want to target for examination can be
indicated here (if you aren't setting the Class field).  If this field is set
after the classinfo object has been initialised, the Objects array will be
refreshed immediately.</p>

###END###

*****************************************************************************/

static ERROR SET_ClassID(struct ClassInfo *Self, LONG *Value)
{
   Self->ClassID = *Value;
   Self->prvClass[0] = 0;

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      acRefresh(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Objects
Short:  Lists all objects by ID number.
Type:   OBJECTID *
Status: Read

###DESCRIPTION###
<p>Following the initialisation of a classinfo object, the Objects field can be
read to get a complete list of objects that exist for the targetted class.
Each object is referenced by its unique ID number.  Both public and private
objects are included in the list, which is sorted by the order of creation (oldest to youngest).
Public objets are easily identifiable by their negative object ID numbers.</p>

<p>The total number of objects listed in the array can be determined from the
TotalObjects field.</p>

###SEE ALSO###
Action: Refresh
Field: TotalObjects

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   TotalObjects
Short:  The total number of objects listed in the Objects array.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>This field indicates the total number of objects that are listed in the
Objects array.</p>

###SEE ALSO###
Field: Objects

###END###

*****************************************************************************/
